# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

"""GBOML version.

The graph-based optimization modeling language (GBOML) version of the language.
"""

__version__ = "0.1.0"
